/* 
 * FILE --- mlp_lm.cc
 *
 * PURPOSE --- Header file for a class that implements a Multilayer 
 * Perceptron (MLP) neural network (NN) with Levenberg-Marquardt (LM) training capability. 
 * The user specifies the number of inputs, the number of outputs, and the number of the 
 * neurons in the hidden layer of the perceptron as well as the number data points that 
 * will be used for training. The traiing algorithm operates in a batch mode, therefore,
 * the user has to provide all the data at the beginning of the training process.
 * It is possible to choose two different activation functions for the neurons that are
 * Log-Sigmoid and Hyperbolic Tangent Sigmoid. The user can either specify the initial 
 * input and output weigth matrices and biaces, or let them be choosen randomly within 
 * user defined range.
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 */


// Include Files
#include <math.h>
#include "mlp_lm.hh"
#include "random.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif


/*************************************************************/

MLP_LM::MLP_LM(int _num_of_inputs, int _num_of_outputs, int _num_of_neurons, int _num_of_data)
{
  int ii, jj;  // Index variales for the loops

  // No error at start
  mlp_lm_err = MLP_LM_NO_ERROR;
  num_of_inputs = _num_of_inputs;
  num_of_outputs = _num_of_outputs;
  num_of_neurons = _num_of_neurons;
  num_of_data = _num_of_data;

  index1 = num_of_inputs*num_of_neurons;
  index2 = index1 + num_of_neurons;
  index3 = index2 + num_of_outputs*num_of_neurons;
  num_of_params = index3 + num_of_outputs;       
 
  // Allocate memory 
  allocate_memory();

  // If any error then report
  if (mlp_lm_err != MLP_LM_NO_ERROR)
    {
#ifdef NO_RCS
      printf("MLP_LM: ERROR --- Not Enought Memory!!!\n");
      printf("The MLP_LM network will not function correctly ...\n");
#else
      rcs_print("MLP_LM: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The MLP_LM network will not function correctly ...\n");
#endif
    }

  // Initialize the gain and bias matrices with random numbers b/n [-1.0, 1.0]
  Randomize();
  set_input_gains(-1.0, 1.0);
  set_input_biases(-1.0, 1.0);
  set_output_gains(-1.0, 1.0);
  set_output_biases(-1.0, 1.0);

  // Set the activation function type as Log-Sigmoid
  set_activation_function_type(LOGSIG);

  // Set the matrix Lambda = 0.5*I as default
  lambda = 0.5;
  Lambda->set_diagonal(lambda);

  // Set the last portion of J (the one related to bo) to identity
  // This portion of J will be constant
  for(ii = 0; ii < num_of_data; ii++)
    {
      for(jj = 0; jj < num_of_outputs; jj++)
	{
	  J->set_element(index3 + jj, 
			 ii*num_of_outputs + jj,
			 1.0);
	}
    }
}
  
MLP_LM::~MLP_LM()
{
  // Free the allocated memory
  free_memory();
}

void MLP_LM::allocate_memory(void)
{
  // Create the matrix of input weigths
  Wi = new matrix(num_of_neurons, num_of_inputs);
  if (Wi->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the vector of input biases
  bi = new matrix(num_of_neurons, 1);
  if (bi->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the matrix of output weigths
  Wo = new matrix(num_of_outputs, num_of_neurons);
  if (Wo->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the vector of output biases
  bo = new matrix(num_of_outputs, 1);
  if (bo->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create input vector u
  u = new matrix(num_of_inputs, 1);
  if (u->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create intermediate vector x
  x = new matrix(num_of_neurons, 1);
  if (x->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create intermediate vector f
  f = new matrix(num_of_neurons, 1);
  if (f->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create output vector y
  y = new matrix(num_of_outputs, 1);
  if (y->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the vector for the parameters
  theta = new matrix(num_of_params, 1);
  if (theta->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the vector for the parameters
  theta_old = new matrix(num_of_params, 1);
  if (theta_old->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the Jacobian matrix
  J = new matrix(num_of_params, num_of_outputs*num_of_data);
  if (J->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create input matrix U
  U = new matrix(num_of_inputs, num_of_data);
  if (U->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create intermediate matrix X
  X = new matrix(num_of_neurons, num_of_data);
  if (X->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create output vector Y
  Y = new matrix(num_of_outputs, num_of_data);
  if (Y->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create output desired vector R
  R = new matrix(num_of_outputs, num_of_data);
  if (R->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the error vector
  E = new matrix(num_of_outputs*num_of_data, 1);
  if (E->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the error vector
  E_old = new matrix(num_of_outputs*num_of_data, 1);
  if (E_old->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create intermediate matrix F
  F = new matrix(num_of_neurons, num_of_data);
  if (F->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the vector del_F
  del_F = new matrix(num_of_neurons, num_of_data);
  if (del_F->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }

  // Create the Lambda matrix
  Lambda = new matrix(num_of_params, num_of_params);
  if (Lambda->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_lm_err = MLP_LM_MEM_ERROR;
    }
}

void MLP_LM::free_memory(void)
{
  delete Wi;
  delete bi;
  delete Wo;
  delete bo;

  delete u;
  delete x;
  delete f;
  delete y;

  delete theta;
  delete theta_old;
  delete J;
  delete U;
  delete X;
  delete Y;
  delete R;
  delete E;
  delete E_old;
  delete F;
  delete del_F;
  delete Lambda;
}


void MLP_LM::set_input_gains(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the input gain matrix to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      for (jj = 0; jj < num_of_inputs; jj++)
	{
	  // Generate a random number b/n [_low, _high]
	  rand_val = Rand_Double(_low, _high);
	  
	  // Assign the element to the generated value
	  Wi->set_element(ii, jj, rand_val);
	}
    }

  // Transfer to the proper portion of theta
  Wi2theta();
}


void MLP_LM::set_input_gains(const double **_in_gains)
{
  // Set the input gain matrix to the specified value
  Wi->set_matrix(_in_gains);

  // Transfer to the proper portion of theta
  Wi2theta();
}


void MLP_LM::set_input_gains(const double *_in_gains)
{
  // Set the input gain matrix to the specified value
  Wi->set_matrix(_in_gains);

  // Transfer to the proper portion of theta
  Wi2theta();
}


void MLP_LM::set_input_biases(double _low, double _high)
{
  int ii;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the input bias vector to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      // Generate a random number b/n [_low, _high]
      rand_val = Rand_Double(_low, _high);
      
      // Assign the element to the generated value
      bi->set_element(ii, 0, rand_val);
    }

  // Transfer to the proper portion of theta
  bi2theta();
}

void MLP_LM::set_input_biases(const double *_in_biases)
{
  // Set the input bias vector to the specified value
  bi->set_matrix(_in_biases);

  // Transfer to the proper portion of theta
  bi2theta();
}


void MLP_LM::set_output_gains(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the output gain matrix to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
	{
	  // Generate a random number b/n [_low, _high]
	  rand_val = Rand_Double(_low, _high);
	  
	  // Assign the element to the generated value
	  Wo->set_element(ii, jj, rand_val);
	}
    }

  // Transfer to the proper portion of theta
  Wo2theta();
}


void MLP_LM::set_output_gains(const double **_out_gains)
{
  // Set the output gain matrix to the specified value
  Wo->set_matrix(_out_gains);

  // Transfer to the proper portion of theta
  Wo2theta();
}


void MLP_LM::set_output_gains(const double *_out_gains)
{
  // Set the output gain matrix to the specified value
  Wo->set_matrix(_out_gains);

  // Transfer to the proper portion of theta
  Wo2theta();
}


void MLP_LM::set_output_biases(double _low, double _high)
{
  int ii;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the output bias vector to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      // Generate a random number b/n [_low, _high]
      rand_val = Rand_Double(_low, _high);
      
      // Assign the element to the generated value
      bo->set_element(ii, 0, rand_val);
    }

  // Transfer to the proper portion of theta
  bo2theta();
}

void MLP_LM::set_output_biases(const double *_out_biases)
{
  // Set the output bias vector to the specified value
  bo->set_matrix(_out_biases);

  // Transfer to the proper portion of theta
  bo2theta();
}

void MLP_LM::set_lambda(double _lambda)
{
  // Make sure that _lambda is positive
  if (_lambda < 0.0)
    {
      lambda = - _lambda;
    }
  else
    {
      lambda = _lambda;
    }

  // Make sure that _lambda bounded
  if (lambda > LAMBDA_UPPER_BOUND)
    {
      lambda = LAMBDA_UPPER_BOUND;
    }
  else if (lambda < LAMBDA_LOWER_BOUND)
    {
      lambda = LAMBDA_LOWER_BOUND;
    }

  // Set the matrix Lambda = lambda*I
  Lambda->set_diagonal(lambda);
}


void MLP_LM::set_activation_function_type(int _actv_func_type)
{
  actv_func_type = _actv_func_type;

  switch(actv_func_type)
    {
    case LOGSIG:  // Log-Sigmoid activation function
      actv_func = logsig;
      del_actv_func = del_logsig;
      break;

    case TANSIG:  // Hyperbolic Tangent Sigmoid activation function
      actv_func = tansig;
      del_actv_func = del_tansig;
      break;

    default:
#ifdef NO_RCS
      printf("Invalid activation function type!!!\n");
      printf("Log-Sigmoid function assumed ...\n");
#else
      rcs_print("Invalid activation function type!!!\n");
      rcs_print("Log-Sigmoid function assumed ...\n");
#endif
      actv_func_type = LOGSIG;
      actv_func = logsig;
    }
}


void MLP_LM::train(const double **input, const double **ref_output, double _tol)
{
  double e, e_old, norm_grad; // Variable for the error

  // Use matrix operations
  U->set_matrix(input);
  R->set_matrix(ref_output);

  do {
    
    // Calculate the weigthed input
    (*X) = (*Wi) * (*U);
    
    // Calculate the output of the activation functions
    get_F();
    
    // Calculate the output of the last (output) layer
    (*Y) = (*Wo) * (*F);
    
    // Add the output biases to each column
    add_out_bias();
    
    // Calculate the error b/n the output and the reference
    get_out_error();
    
    // Calculate the sum of the squared errors
    e_old = (transpose((*E)) * (*E)).get_element(0, 0);

    // Calculate the derivative of the activation functions
    get_del_F();
    
    // Calculate the Jacobian
    get_J();

    // Calculate the norm of the gradient
    //    norm_grad = norm((*J) * (*E));
    norm_grad = get_norm_grad();

    // Save the previos vector of parameters
    (*theta_old) = (*theta);
    (*E_old) = (*E);

    do {

      // Update the parameters
      (*theta) = (*theta) - inv((*J) * transpose((*J)) + (*Lambda)) * (*J) * (*E);

      // Transfer the updated parameters
      theta2Wi();
      theta2bi();
      theta2Wo();
      theta2bo();
      
      // Calculate the weigthed input
      (*X) = (*Wi) * (*U);
      
      // Calculate the output of the activation functions
      get_F();
      
      // Calculate the output of the last (output) layer
      (*Y) = (*Wo) * (*F);
      
      // Add the output biases to each column
      add_out_bias();
      
      // Calculate the error b/n the output and the reference
      get_out_error();
      
      // Calculate the sum of the squared errors
      e = (transpose(*E) * (*E)).get_element(0, 0);

      // Restore the old parameters and increse lambda if no improvement
      if (e > e_old)
	{
	  // Decrease the step size (since the inverse of Lambda)
	  set_lambda(10.0 * lambda);

	  // Restore the previos vector of parameters
	  (*theta) = (*theta_old);
	  (*E) = (*E_old);
	}
    } while (e > e_old);
    
    // Increase the step size (since the inverse of Lambda)
    set_lambda(0.1 * lambda);
    
  } while ((e > _tol) & (norm_grad > NORM_LOWER_BOUND));

  // Print the final sum of the errors
#ifdef NO_RCS
    printf("Sum of the errors: %f\n", e);
#else
    rcs_print("Sum of the errors: %f\n", e);
#endif
}

double MLP_LM::train_one_step(const double **input, const double **ref_output)
{
  double e, e_old, norm_grad; // Variable for the error

  // Use matrix operations
  U->set_matrix(input);
  R->set_matrix(ref_output);

  // Calculate the weigthed input
  (*X) = (*Wi) * (*U);
  
  // Calculate the output of the activation functions
  get_F();
  
  // Calculate the output of the last (output) layer
  (*Y) = (*Wo) * (*F);
  
  // Add the output biases to each column
  add_out_bias();
  
  // Calculate the error b/n the output and the reference
  get_out_error();
  
  // Calculate the sum of the squared errors
  e_old = (transpose((*E)) * (*E)).get_element(0, 0);

  // Calculate the derivative of the activation functions
  get_del_F();
  
  // Calculate the Jacobian
  get_J();
  
  // Calculate the norm of the gradient
  //  norm_grad = norm((*J) * (*E));
  norm_grad = get_norm_grad();

  // Save the previos vector of parameters
  (*theta_old) = (*theta);
  (*E_old) = (*E);
  
  do {
    
    // Update the parameters
    (*theta) = (*theta) - inv((*J) * transpose((*J)) + (*Lambda)) * (*J) * (*E);
    
    // Transfer the updated parameters
    theta2Wi();
    theta2bi();
    theta2Wo();
    theta2bo();
    
    // Calculate the weigthed input
    (*X) = (*Wi) * (*U);
    
    // Calculate the output of the activation functions
    get_F();
    
    // Calculate the output of the last (output) layer
    (*Y) = (*Wo) * (*F);
    
    // Add the output biases to each column
    add_out_bias();
    
    // Calculate the error b/n the output and the reference
    get_out_error();
    
    // Calculate the sum of the squared errors
    e = (transpose(*E) * (*E)).get_element(0, 0);

    // Restore the old parameters and increse lambda if no improvement
    if (e > e_old)
      {
	// Decrease the step size (since the inverse of Lambda)
	set_lambda(10.0 * lambda);
	
	// Restore the previos vector of parameters
	(*theta) = (*theta_old);
	(*E) = (*E_old);
      }
  } while (e > e_old);
    
  // Increase the step size (since the inverse of Lambda)
  set_lambda(0.1 * lambda);
}

void MLP_LM::test(const double *input, double *_output)
{
  // Use matrix operations
  u->set_col(0, input);

  // Calculate the output of the input layer
  (*x) = (*Wi) * (*u) + (*bi);

  // Calculate the output of the activation functions
  get_activation();

  // Calculate the output of the output layer
  (*y) = (*Wo) * (*f) + (*bo);
  
  // Return the output
  get_output(_output);
}

void MLP_LM::Wi2theta(void)
{
  int ii, jj;  // Index variales for the loops

  // Set the input gain matrix portion of theta
  for (ii=0; ii < num_of_neurons; ii++)
    {
      for (jj=0; jj < num_of_inputs; jj++)
	{
	  // Assign the element from the vector of parameters
	  theta->set_element(ii*num_of_inputs + jj, 0, Wi->get_element(ii, jj));
	}
    }
}


void MLP_LM::bi2theta(void)
{
  int ii;  // Index variales for the loops

  // Set the input bias vector portion of theta
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      // Assign the element to the generated value
      theta->set_element(index1 + ii, 0, bi->get_element(ii, 0));
    }
}


void MLP_LM::Wo2theta(void)
{
  int ii, jj;  // Index variales for the loops

  // Set the output gain matrix portion of theta
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
	{
	  // Assign the element to the generated value
	  theta->set_element(index2 + ii*num_of_neurons + jj, 0, Wo->get_element(ii, jj));
	}
    }
}


void MLP_LM::bo2theta(void)
{
  int ii;  // Index variales for the loops

  // Set the output bias portion of theta
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      // Assign the element to the generated value
      theta->set_element(index3 + ii, 0, bo->get_element(ii, 0));
    }
}


void MLP_LM::theta2Wi(void)
{
  int ii, jj;  // Index variales for the loops

  // Set the input gain matrix to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      for (jj = 0; jj < num_of_inputs; jj++)
	{
	  // Assign the element from the vector of parameters
	  Wi->set_element(ii, jj, theta->get_element(ii*num_of_inputs + jj, 0));
	}
    }
}


void MLP_LM::theta2bi(void)
{
  int ii;  // Index variales for the loops

  // Set the input bias vector to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      // Assign the element to the generated value
      bi->set_element(ii, 0, theta->get_element(index1 + ii, 0));
    }
}


void MLP_LM::theta2Wo(void)
{
  int ii, jj;  // Index variales for the loops

  // Set the output gain matrix to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
	{
	  // Assign the element to the generated value
	  Wo->set_element(ii, jj, 
			  theta->get_element(index2 + ii*num_of_neurons + jj, 0));
	}
    }
}


void MLP_LM::theta2bo(void)
{
  int ii;  // Index variales for the loops

  // Set the output bias vector to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      // Assign the element to the generated value
      bo->set_element(ii, 0, theta->get_element(index3 + ii, 0));
    }
}


void MLP_LM::get_activation(void)
{
  double func_val;  // Temporary variable

  // Calculate the activation of each neuron
  for(int ii = 0; ii < num_of_neurons; ii++)
    {
      func_val = actv_func(x->get_element(ii, 0));
      f->set_element(ii, 0, func_val);
    }
}


void MLP_LM::get_F(void)
{
  int ii, jj;  // Index variales for the loops
  double func_val;  // Temporary variable

  // Calculate the activation of each neuron
  for(ii = 0; ii < num_of_neurons; ii++)
    {
      for(jj = 0; jj < num_of_data; jj++)
	{
	  func_val = actv_func(X->get_element(ii, jj) + bi->get_element(ii, 0));
	  F->set_element(ii, jj, func_val);
	}
    }
}

void MLP_LM::add_out_bias(void)
{
  int ii, jj;  // Index variales for the loops

  // Add the output bias vector to each data
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_data; jj++)
	{
	  // Add the corresponding bias value
	  Y->set_element(ii, jj,
			 Y->get_element(ii, jj) + bo->get_element(ii, 0));
	}
    }
}


void MLP_LM::get_del_F(void)
{
  int ii, jj;  // Index variales for the loops
  double func_val;  // Temporary variable

  // Calculate the activation of each neuron
  for(ii = 0; ii < num_of_neurons; ii++)
    {
      for(jj = 0; jj < num_of_data; jj++)
	{
	  func_val = del_actv_func(X->get_element(ii, jj) + bi->get_element(ii, 0));
	  del_F->set_element(ii, jj, func_val);
	}
    }
}

void MLP_LM::get_J(void)
{
  int ii, jj, kk, ll;  // Index variales for the loops
  double temp[num_of_neurons];  // Temporary variable

  
  for(ii = 0; ii < num_of_data; ii++)
    {
      for(jj = 0; jj < num_of_outputs; jj++)
	{
	  // Calculate a common variale in the derivative wrt the input parameters
	  for(kk = 0; kk < num_of_neurons; kk++)
	    {
	      temp[kk] = Wo->get_element(jj, kk)*del_F->get_element(kk, ii);
	    }
	  
	  // The portion of J due to Wi
	  for(kk = 0; kk < num_of_neurons; kk++)
	    {
	      for(ll = 0; ll < num_of_inputs; ll++)
		{
		  J->set_element(kk*num_of_inputs + ll, 
				 ii*num_of_outputs + jj,
				 temp[kk]*U->get_element(ll, ii));
		}
	    }
	  
	  // The portion of J due to bi
	  for(kk = 0; kk < num_of_neurons; kk++)
	    {
	      J->set_element(index1 + kk, 
			     ii*num_of_outputs + jj, 
			     temp[kk]);
	    }
	  
	  // The portion of J due to Wo
	  for(kk = 0; kk < num_of_neurons; kk++)
	    {
	      J->set_element(index2 + jj*num_of_neurons + kk, 
			     ii*num_of_outputs + jj,
			     F->get_element(kk, ii));
	    }
	  
	  // The portion of J due to bo is always identity matrix 
	  // and does not need to be updated
	  // ...
	}
    }
}


void MLP_LM::get_output(double *_output)
{
  // Return the output
  for(int ii = 0; ii < num_of_outputs; ii++)
    {
      _output[ii] = y->get_element(ii, 0);
    }
}

 
void MLP_LM::get_out_error(void)
{
  int ii, jj;  // Index variales for the loops

  // Calculate the vector of output errors 
  for (ii = 0; ii < num_of_data; ii++)
    {
      for (jj = 0; jj < num_of_outputs; jj++)
	{
	  E->set_element(ii*num_of_outputs + jj, 0,
			 Y->get_element(jj, ii) - R->get_element(jj, ii));
	}
    }
}

inline double MLP_LM::get_norm_grad(void)
{
  return norm((*J) * (*E));
}


// Activation Functions and their derivatives wrt an input parameter

inline double logsig(const double _input)
{
  return ( (1.0 / (1.0 + exp(-_input))) );
}

inline double del_logsig(const double _input)
{
  return ( (exp(-_input) / ((1.0 + exp(-_input))*(1.0 + exp(-_input)))) );
}


inline double tansig(const double _input)
{
  return ( ((exp(_input) - exp(-_input)) / (exp(_input) + exp(-_input))) );
}

inline double del_tansig(const double _input)
{
  return ( (4.0 / ((exp(_input) + exp(-_input))*(exp(_input) + exp(-_input)))) );
}

